import gleam/dynamic/decode.{type Decoder}
import gleam/json
import gleam/list
import gleam/option.{type Option}
import models/construction.{type Construction}
import models/jump_gate.{type JumpGate}
import models/market.{type Market}
import models/ship_cargo.{type ShipCargo}
import models/ship_symbol.{type ShipSymbol}
import models/shipyard.{type Shipyard}
import models/system.{type System}
import models/system_symbol.{type SystemSymbol}
import models/trade_symbol.{type TradeSymbol}
import models/waypoint.{type Waypoint}
import models/waypoint_symbol.{type WaypointSymbol}
import models/waypoint_trait_symbol.{type WaypointTraitSymbol}
import models/waypoint_type.{type WaypointType}
import utils/api.{type ApiResponse, type PagedData}
import utils/auth.{type AgentToken, AgentAuth}

pub type ListSystemsResponse {
  ListSystemsResponse(systems: List(System))
}

fn list_systems_response_decoder() -> Decoder(ListSystemsResponse) {
  use systems <- decode.then(decode.list(system.decoder()))
  decode.success(ListSystemsResponse(systems:))
}

pub fn list_systems(
  token: AgentToken,
  page: Int,
  limit: Int,
) -> ApiResponse(PagedData(ListSystemsResponse)) {
  let request = api.get_page(AgentAuth(token), "/systems", page, limit)
  use response <- api.try_send(request)
  case response.status {
    200 ->
      api.parse_paged_data_response(response, list_systems_response_decoder())
    _ -> api.parse_error_response(response)
  }
}

pub type GetSystemResponse {
  GetSystemResponse(system: System)
}

fn get_system_response_decoder() -> Decoder(GetSystemResponse) {
  use system <- decode.then(system.decoder())
  decode.success(GetSystemResponse(system:))
}

pub fn get_system(
  token: AgentToken,
  system_symbol: SystemSymbol,
) -> ApiResponse(GetSystemResponse) {
  let request =
    api.get(
      AgentAuth(token),
      "/systems/" <> system_symbol.to_string(system_symbol),
    )
  use response <- api.try_send(request)
  case response.status {
    200 -> api.parse_data_response(response, get_system_response_decoder())
    _ -> api.parse_error_response(response)
  }
}

pub type ListSystemWaypointsResponse {
  ListSystemWaypointsResponse(waypoints: List(Waypoint))
}

fn list_system_waypoints_response_decoder() -> Decoder(
  ListSystemWaypointsResponse,
) {
  use waypoints <- decode.then(decode.list(waypoint.decoder()))
  decode.success(ListSystemWaypointsResponse(waypoints:))
}

pub fn list_system_waypoints(
  token: AgentToken,
  system_symbol: SystemSymbol,
  page: Int,
  limit: Int,
  type_: Option(WaypointType),
  traits: List(WaypointTraitSymbol),
) -> ApiResponse(PagedData(ListSystemWaypointsResponse)) {
  let traits_query =
    traits
    |> list.map(fn(trait) {
      #("traits", waypoint_trait_symbol.to_string(trait))
    })
  let query = case type_ {
    option.Some(type_) -> [
      #("type", waypoint_type.to_string(type_)),
      ..traits_query
    ]
    option.None -> traits_query
  }
  let request =
    api.get_page_query(
      AgentAuth(token),
      "/systems/" <> system_symbol.to_string(system_symbol) <> "/waypoints",
      page,
      limit,
      query,
    )
  use response <- api.try_send(request)
  case response.status {
    200 ->
      api.parse_paged_data_response(
        response,
        list_system_waypoints_response_decoder(),
      )
    _ -> api.parse_error_response(response)
  }
}

pub type GetWaypointResponse {
  GetWaypointResponse(waypoint: Waypoint)
}

fn get_waypoint_response_decoder() -> Decoder(GetWaypointResponse) {
  use waypoint <- decode.then(waypoint.decoder())
  decode.success(GetWaypointResponse(waypoint:))
}

pub fn get_waypoint(
  token: AgentToken,
  system_symbol: SystemSymbol,
  waypoint_symbol: WaypointSymbol,
) -> ApiResponse(GetWaypointResponse) {
  let request =
    api.get(
      AgentAuth(token),
      "/systems/"
        <> system_symbol.to_string(system_symbol)
        <> "/waypoints/"
        <> waypoint_symbol.to_string(waypoint_symbol),
    )
  use response <- api.try_send(request)
  case response.status {
    200 -> api.parse_data_response(response, get_waypoint_response_decoder())
    _ -> api.parse_error_response(response)
  }
}

pub type GetConstructionSiteResponse {
  GetConstructionSiteResponse(construction: Construction)
}

fn get_construction_site_response_decoder() -> Decoder(
  GetConstructionSiteResponse,
) {
  use construction <- decode.then(construction.decoder())
  decode.success(GetConstructionSiteResponse(construction:))
}

pub fn get_construction_site(
  token: AgentToken,
  system_symbol: SystemSymbol,
  waypoint_symbol: WaypointSymbol,
) -> ApiResponse(GetConstructionSiteResponse) {
  let request =
    api.get(
      AgentAuth(token),
      "/systems/"
        <> system_symbol.to_string(system_symbol)
        <> "/waypoints/"
        <> waypoint_symbol.to_string(waypoint_symbol)
        <> "/construction",
    )
  use response <- api.try_send(request)
  case response.status {
    200 ->
      api.parse_data_response(
        response,
        get_construction_site_response_decoder(),
      )
    _ -> api.parse_error_response(response)
  }
}

pub type SupplyConstructionSiteResponse {
  SupplyConstructionSiteResponse(construction: Construction, cargo: ShipCargo)
}

fn supply_construction_site_response_decoder() -> Decoder(
  SupplyConstructionSiteResponse,
) {
  use construction <- decode.field("construction", construction.decoder())
  use cargo <- decode.field("cargo", ship_cargo.decoder())
  decode.success(SupplyConstructionSiteResponse(construction:, cargo:))
}

pub fn supply_construction_site(
  token: AgentToken,
  system_symbol: SystemSymbol,
  waypoint_symbol: WaypointSymbol,
  ship_symbol: ShipSymbol,
  trade_symbol: TradeSymbol,
  units: Int,
) -> ApiResponse(SupplyConstructionSiteResponse) {
  let request =
    api.post_json(
      AgentAuth(token),
      "/systems/"
        <> system_symbol.to_string(system_symbol)
        <> "/waypoints/"
        <> waypoint_symbol.to_string(waypoint_symbol)
        <> "/construction/supply",
      json.object([
        #("shipSymbol", ship_symbol.encode(ship_symbol)),
        #("tradeSymbol", trade_symbol.encode(trade_symbol)),
        #("units", json.int(units)),
      ]),
    )
  use response <- api.try_send(request)
  case response.status {
    201 ->
      api.parse_data_response(
        response,
        supply_construction_site_response_decoder(),
      )
    _ -> api.parse_error_response(response)
  }
}

pub type GetMarketResponse {
  GetMarketResponse(market: Market)
}

fn get_market_response_decoder() -> Decoder(GetMarketResponse) {
  use market <- decode.then(market.decoder())
  decode.success(GetMarketResponse(market:))
}

pub fn get_market(
  token: AgentToken,
  system_symbol: SystemSymbol,
  waypoint_symbol: WaypointSymbol,
) -> ApiResponse(GetMarketResponse) {
  let request =
    api.get(
      AgentAuth(token),
      "/systems/"
        <> system_symbol.to_string(system_symbol)
        <> "/waypoints/"
        <> waypoint_symbol.to_string(waypoint_symbol)
        <> "/market",
    )
  use response <- api.try_send(request)
  case response.status {
    200 -> api.parse_data_response(response, get_market_response_decoder())
    _ -> api.parse_error_response(response)
  }
}

pub type GetJumpGateResponse {
  GetJumpGateResponse(jump_gate: JumpGate)
}

fn get_jump_gate_response_decoder() -> Decoder(GetJumpGateResponse) {
  use jump_gate <- decode.then(jump_gate.decoder())
  decode.success(GetJumpGateResponse(jump_gate:))
}

pub fn get_jump_gate(
  token: AgentToken,
  system_symbol: SystemSymbol,
  waypoint_symbol: WaypointSymbol,
) -> ApiResponse(GetJumpGateResponse) {
  let request =
    api.get(
      AgentAuth(token),
      "/systems/"
        <> system_symbol.to_string(system_symbol)
        <> "/waypoints/"
        <> waypoint_symbol.to_string(waypoint_symbol)
        <> "/jump-gate",
    )
  use response <- api.try_send(request)
  case response.status {
    200 -> api.parse_data_response(response, get_jump_gate_response_decoder())
    _ -> api.parse_error_response(response)
  }
}

pub type GetShipyardResponse {
  GetShipyardResponse(shipyard: Shipyard)
}

fn get_shipyard_response_decoder() -> Decoder(GetShipyardResponse) {
  use shipyard <- decode.then(shipyard.decoder())
  decode.success(GetShipyardResponse(shipyard:))
}

pub fn get_shipyard(
  token,
  system_symbol,
  waypoint_symbol,
) -> ApiResponse(GetShipyardResponse) {
  let request =
    api.get(
      AgentAuth(token),
      "/systems/"
        <> system_symbol
        <> "/waypoints/"
        <> waypoint_symbol
        <> "/shipyard",
    )
  use response <- api.try_send(request)
  case response.status {
    200 -> api.parse_data_response(response, get_shipyard_response_decoder())
    _ -> api.parse_error_response(response)
  }
}
